
const db = require('machankura.db');
const reggie = require("machankura.lib/reggie")
const bolt11 = require("bolt11")
const lightningNodeUtil = require("machankura.lib/lightning-node-util")
const idGenerator = require("machankura.db/src/lib/id-generator")
const machankuraUserQueries = require("machankura.lib/queries/user")

const machankuraDestinationParameter = (destination = "Bolt11") => {
    return destination
}

module.exports.sendBitcoinToBolt11Invoice = async (request, response, next) => {
    if (request.platform) {
        if (request.body.bolt11.match(reggie.bolt11Regex)) {
            const bolt11Object = bolt11.decode(request.body.bolt11)
            
            // TODO: Get lightning address from the description
            if (bolt11Object) {
                const t = await db.sequelize.transaction();
                const satsBalance = await request.platform.getUserBalance(db, t)
                const destinationParameter = machankuraDestinationParameter(request.body.destination)

                const invoicedSatoshis = Math.ceil((bolt11Object.millisatoshis ?? 0)/1_000)
                const satoshisBeingSent = Number(request.body.amountInSatoshis ?? invoicedSatoshis)

                if (satoshisBeingSent < invoicedSatoshis) {
                    response.status(400)
                    return response.json({
                        status: "error",
                        message: "We can't send less then we were invoiced"
                    })
                } else if (satsBalance > satoshisBeingSent) {
                    // Submit a request to send Bitcoin...
                    // TODO: Check if invoice belongs to machankura user...
                    // TODO: Decode bolt11Invoice to get expiry...
                    const transfer = await db.Transfer.create({
                        status: "pending",
                        sourceId: request.platform.id,
                        amountInSatoshis: satoshisBeingSent,
                        feeInSatoshis: Math.ceil(satoshisBeingSent*1/100),
                        destinationParameter: destinationParameter, // TODO: Might be lightning address...
                        bolt11Transfer: {
                            id: idGenerator.generateUniqueId(), // Should be unique...
                            invoice: request.body.bolt11,
                            status: "pending",
                            amountInSatoshis: satoshisBeingSent,
                            expiresAt: new Date(bolt11Object.timeExpireDateString), // TODO: Set this from the bolt11 invoice
                        }
                    }, {
                        include: [
                            {
                                association: db.Transfer.Bolt11Transfer
                            }
                        ],
                        transaction: t
                    }).catch(error => {
                        console.error("Failed to send sats: ", error)
                        return null
                    })
                    
                    if (transfer) {
                        await t.commit()
                        response.status(201)
                        return response.json({
                            status: "ok",
                            message: "transfer initiatilized",
                            data: transfer.toJSON()
                        })
                    } else {
                        response.status(500)
                    }
                } else {
                    response.status(413)
                    response.json({
                        status: "error",
                        message: "Insufficient balance"
                    })
                }
                
                await t.rollback();
            } else {
                response.status(400)
                response.json({
                    status: "error",
                    message: "Invalid bolt11 invoice."
                })
            }
            
        } else {
            response.status(400)
            response.json({
                status: "error",
                message: "Invalid bolt11 invoice."
            })
        }
    } else {
        response.status(401)
        response.json({
            status: "error",
            message: "Unauthorized"
        })
    }
}

module.exports.sendBitcoinToOnchain = async (request, response, next) => {
    if (request.platform) {
        const destinationParameter = request.body.destinationParameter

        if (request.body.destinationParameter.match(reggie.bitcoinAddressRegex)) {

            const t = await db.sequelize.transaction();
            const satsBalance = await request.platform.getUserBalance(db, t)
            const amountInSatoshis = Number(request.body.amountInSatoshis)
            const serviceFee = Math.ceil(amountInSatoshis*1/100);

            
            if (amountInSatoshis <= satsBalance) {

                const onchainBitcoin = await db.OnchainBitcoin.findByPk(
                    destinationParameter,
                    {
                        include: [
                            {
                                association: db.OnchainBitcoin.UserOnchainBitcoin,
                                required: true
                            }
                        ]
                    }
                )
                
                if (onchainBitcoin) {
                    // Machankura transfer
                    if (onchainBitcoin.userOnchainBitcoin.userId == request.platform.id) {
                        await t.rollback();
                        response.status(400)
                        return response.json({
                            status: "error",
                            message: "You can't send Bitcoin to yourself."
                        })
                    } else {
                        const transfer = await db.Transfer.create({
                            sourceId: request.platform.id,
                            destinationId: onchainBitcoin.userOnchainBitcoin.userId,
                            amountInSatoshis: amountInSatoshis, // TODO: Convert amount to satoshis just in case...
                            feeInSatoshis: serviceFee,
                            destinationParameter,
                            status: "pending"
                        }).catch(e => {
                            console.error("Failed to create transfer: ", e)
                            return null
                        })
                        
                        // const updatedBalance = await request.platform.getUserBalance(db)
                        if (transfer) {
                            await t.commit();
                            // await notifier.notifyUserBySMS(
                            //     db,
                            //     onchainBitcoin.userOnchainBitcoin.userId,
                            //     amountInSatoshis,
                            //     getUserName(request.platform)
                            // )
                            
                            response.status(201)
                            return response.json({
                                status: "ok",
                                message: "transfer initiatilized",
                                data: transfer.toJSON()
                            })
                        } else {
                            await t.rollback();
                            
                            response.status(400)
                            return response.json({
                                status: "error",
                                message: "Failed to create transfer."
                            })
                        }
                    }
                } else {
                    const transfer = await db.Transfer.create({
                        sourceId: request.platform.id,
                        amountInSatoshis: amountInSatoshis,
                        feeInSatoshis: serviceFee,
                        destinationParameter,
                        status: "pending"
                    }, {
                        transaction: t
                    }).catch(e => {
                        console.error("Failed to create transfer: ", e)
                        return null
                    })
    
                    if (transfer) {

                        await t.commit();
                        response.status(201)
                        return response.json({
                            status: "ok",
                            message: "transfer initiatilized",
                            data: transfer.toJSON()
                        })
                    } else {
                        await t.rollback();
                        return `END Failed to initiate onchain Bitcoin transaction.`
                    }
                }
            } else {
                response.status(413)
                response.json({
                    status: "error",
                    message: "Insufficient balance"
                })
            }
            
            await t.rollback();
            
        } else {
            response.status(400)
            return response.json({
                status: "error",
                message: "Invalid onchain address."
            })
        }
    } else {
        response.status(401)
        return response.json({
            status: "error",
            message: "Unauthorized"
        })
    }
}

const getReceiver = async (db, destinationParameter, transferReceiver = null) => {
    if (transferReceiver) {
        return transferReceiver
    } else {
        return await machankuraUserQueries.getUser(
            db,
            destinationParameter
        )
    }
}
module.exports.sendBitcoinToMachankuraUser = async (request, response, next, transferReceiver = null) => {
    if (request.platform) {
        const destinationParameter = request.body.destinationParameter.replace("+","")

        const receiverUser = await getReceiver(
            db,
            destinationParameter,
            transferReceiver
        )
        
        if (receiverUser) {
            const t = await db.sequelize.transaction();
            const satsBalance = await request.platform.getUserBalance(db, t)
            const amountInSatoshis = Number(request.body.amountInSatoshis)
            const serviceFee = 0;

            if (amountInSatoshis <= satsBalance) {

                // Machankura transfer
                if (receiverUser.id == request.platform.id) {
                    await t.rollback();
                    response.status(400)
                    return response.json({
                        status: "error",
                        message: "You can't send Bitcoin to yourself."
                    })
                } else {
                    const transfer = await db.Transfer.create({
                        sourceId: request.platform.id,
                        destinationId: receiverUser.id,
                        amountInSatoshis: amountInSatoshis, // TODO: Convert amount to satoshis just in case...
                        feeInSatoshis: serviceFee,
                        destinationParameter,
                        status: "pending"
                    }).catch(e => {
                        console.error("Failed to create transfer: ", e)
                        return null
                    })
                    
                    // const updatedBalance = await request.platform.getUserBalance(db)
                    if (transfer) {
                        await t.commit();
                        // await notifier.notifyUserBySMS(
                        //     db,
                        //     receiverUser.id,
                        //     amountInSatoshis,
                        //     getUserName(request.platform)
                        // )
                        
                        response.status(201)
                        return response.json({
                            status: "ok",
                            message: "user transfer initiatilized",
                            data: transfer.toJSON()
                        })
                    } else {
                        await t.rollback();
                        
                        response.status(400)
                        return response.json({
                            status: "error",
                            message: "Failed to create transfer."
                        })
                    }
                }
            } else {
                response.status(413)
                response.json({
                    status: "error",
                    message: "Insufficient balance"
                })
            }
            
            await t.rollback();
        } else {
            // TODO: Consider putting gifting logic here 
            response.status(400)
            response.json({
                status: "error",
                message: "Invalid user identifier."
            })
        }
    } else {
        response.status(401)
        response.json({
            status: "error",
            message: "Unauthorized"
        })
    }
}

module.exports.giftBitcoinToPhoneNumber = async (request, response, next) => {
    if (request.platform) {
        const destinationParameter = request.body.destinationParameter.replace("+", "")

        const receiverUser = machankuraUserQueries.getUser(
            db,
            destinationParameter
        )
        
        if (receiverUser == null) {
            const t = await db.sequelize.transaction();
            const satsBalance = await request.platform.getUserBalance(db, t)
            const amountInSatoshis = Number(request.body.amountInSatoshis)
            const serviceFee = 0;

            if (amountInSatoshis <= satsBalance) {

                const transfer = await db.Transfer.create({
                    sourceId: request.platform.id,
                    // destinationId: receiverUser.id,
                    amountInSatoshis: amountInSatoshis, // TODO: Convert amount to satoshis just in case...
                    feeInSatoshis: serviceFee,
                    destinationParameter,
                    status: "gift"
                }).catch(e => {
                    console.error("Failed to create transfer: ", e)
                    return null
                })
                
                // const updatedBalance = await request.platform.getUserBalance(db)
                if (transfer) {
                    await t.commit();
                    // await notifier.notifyUserBySMS(
                    //     db,
                    //     receiverUser.id,
                    //     amountInSatoshis,
                    //     getUserName(request.platform)
                    // )
                    
                    response.status(201)
                    return response.json({
                        status: "ok",
                        message: "transfer initiatilized",
                        data: transfer.toJSON()
                    })
                } else {
                    await t.rollback();
                    
                    response.status(400)
                    return response.json({
                        status: "error",
                        message: "Failed to create transfer."
                    })
                }
            } else {
                response.status(413)
                response.json({
                    status: "error",
                    message: "Insufficient balance"
                })
            }
            
            await t.rollback();
        } else {
            response.status(400)
            response.json({
                status: "error",
                message: "Invalid onchain address."
            })
        }
    } else {
        response.status(401)
        response.json({
            status: "error",
            message: "Unauthorized"
        })
    }
}

module.exports.sendBitcoinToLightningAddress = async (request, response, next, transferReceiver = null) => {
    if (request.platform) {
        const destinationParameter = request.body.destinationParameter

        if (destinationParameter.endsWith("@8333.mobi")) {
            await this.sendBitcoinToMachankuraUser(
                request,
                response,
                next,
                transferReceiver
            )
        } else {
            // TODO: Send Bitcoin to external lightning address...
            if (request.body.destinationParameter.match(reggie.lightningAddressRegex)) {

                const t = await db.sequelize.transaction();
                const satsBalance = await request.platform.getUserBalance(db, t)
                const amountInSatoshis = Number(request.body.amountInSatoshis)
                const serviceFee = Math.ceil(amountInSatoshis*1/100);
    
                
                if (amountInSatoshis <= satsBalance) {
                    const transfer = await db.Transfer.create({
                        sourceId: request.platform.id,
                        amountInSatoshis: amountInSatoshis,
                        feeInSatoshis: serviceFee,
                        destinationParameter,
                        status: "pending"
                    }, {
                        transaction: t
                    }).catch(e => {
                        console.error("Failed to create transfer: ", e)
                        return null
                    })
    
                    if (transfer) {

                        await t.commit();
                        response.status(201)
                        return response.json({
                            status: "ok",
                            message: "transfer initiatilized",
                            data: transfer.toJSON()
                        })
                    } else {
                        await t.rollback();
                        return `END Failed to initiate onchain Bitcoin transaction.`
                    }
                } else {
                    response.status(413)
                    response.json({
                        status: "error",
                        message: "Insufficient balance"
                    })
                }
                
                await t.rollback();
                
            } else {
                response.status(400)
                return response.json({
                    status: "error",
                    message: "Invalid lightning address."
                })
            }
        }        
    } else {
        response.status(401)
        response.json({
            status: "error",
            message: "Unauthorized"
        })
    }
}