const { expect } = require('@jest/globals');
const db = require('machankura.db');
const sendBitcoinMiddleware = require("../../server/handler/send")
const mocked = require('../utils/mocked')

beforeEach(async () => {
    await db.init('memory');
});

afterEach(async () => {
    await db.sequelize.close();
});
const pin = "55555"

const alicePhoneNumber = "27123456789"
const bobPhoneNumber = "27987699321"
const bobPhoneNumberStartingWithZero = bobPhoneNumber.replace("27", "0")
const carolPhoneNumber = "27444499322"
const carolLightningAddress = "carol@8333.mobi"

const next = () => {}

const init = async () => {
    const country = await db.Country.create({
        id: "ZA", 
        name: "South Africa",
        countryCallingCode: "27",
        currencyTicker: "ZAR",
        serviceCountry: {
            status: "GENERALLY_AVAILABLE",
            africasTalking: {
                serviceCode: "*8333#",
                apiKey: "randomlygeneratedapikey",
                username: "at_country_username"
            },
            ussdServices: [
                {
                    serviceCode: "*8333#"
                }
            ]
        }
    }, {
        include: [
            {
                association: db.Country.ServiceCountry,
                include: [
                    {
                        association: db.ServiceCountry.AfricasTalking
                    },
                    {
                        association: db.ServiceCountry.UssdServices
                    }
                ]
            }
        ]
    })
    const alice = await db.User.create({
        userPins: [
            {
                pin: {
                    code: pin
                }
            }
        ],
        userPhones: [
            {
                phone: {
                    number: alicePhoneNumber
                }
            }
        ]
    }, {
        include: [
            {
                association: db.User.UserPins,
                include: [
                    {
                        association: db.UserPin.Pin
                    }
                ]
            },
            {
                association: db.User.UserPhones,
                include: [
                    {
                        association: db.UserPhone.Phone
                    }
                ]
            }
        ]
    })

    const bob = await db.User.create({
        userPins: [
            {
                pin: {
                    code: pin
                }
            }
        ],
        userPhones: [
            {
                phone: {
                    number: bobPhoneNumber
                }
            }
        ]
    }, {
        include: [
            {
                association: db.User.UserPins,
                include: [
                    {
                        association: db.UserPin.Pin
                    }
                ]
            },
            {
                association: db.User.UserPhones,
                include: [
                    {
                        association: db.UserPhone.Phone
                    }
                ]
            }
        ]
    })

    const carol = await db.User.create({
        userPins: [
            {
                pin: {
                    code: pin
                }
            }
        ],
        userPhones: [
            {
                phone: {
                    number: carolPhoneNumber
                }
            }
        ],
        userName: {
            name: {
                text: "carol"
            }
        }
    }, {
        include: [
            {
                association: db.User.UserPins,
                include: [
                    {
                        association: db.UserPin.Pin
                    }
                ]
            },
            {
                association: db.User.UserPhones,
                include: [
                    {
                        association: db.UserPhone.Phone
                    }
                ]
            },
            {
                association: db.User.UserName,
                include: [
                    {
                        association: db.UserName.Name
                    }
                ]
            }
        ]
    })

    
    const amount = 2000
    const lightningAddressToAlice = await db.Transfer.create({
        destinationId: alice.id,
        amountInSatoshis: amount,
        feeInSatoshis: Math.ceil(amount*1/100),
        destinationParameter: `${alicePhoneNumber}@8333.mobi`,
        status: "settled"
    })
    
    return {
        country,
        alice,
        bob,
        carol
    }
}

test("Send Bitcoin to machankura username lightning address", async () => {
    const content = await init();

    const request = {
        user: content.alice,
        body: {
            destinationParameter: carolLightningAddress,
            amountInSatoshis: 21,
            description: "sending to username"
        }
    }
    const sendBitcoinResponse = mocked.mockResponse()

    await sendBitcoinMiddleware.sendBitcoinToLightningAddress(
        request,
        sendBitcoinResponse,
        next
    )
    
    expect(sendBitcoinResponse.statusCode).toBe(201)
    expect(sendBitcoinResponse.jsonResponse.status).toBe("ok")
    expect(sendBitcoinResponse.jsonResponse.data.status).toBe("pending")
    expect(sendBitcoinResponse.jsonResponse.data.amountInSatoshis).toBe(21)
    expect(sendBitcoinResponse.jsonResponse.data.feeInSatoshis).toBe(0)
    expect(sendBitcoinResponse.jsonResponse.data.sourceId).toBe(content.alice.id)
    expect(sendBitcoinResponse.jsonResponse.data.destinationId).toBe(content.carol.id)
    expect(sendBitcoinResponse.jsonResponse.data.destinationParameter).toBe("carol@8333.mobi")
});

test("Send Bitcoin to machankura phoneNumber lightning address", async () => {
    const content = await init();

    const request = {
        user: content.alice,
        body: {
            destinationParameter: `${bobPhoneNumber}@8333.mobi`,
            amountInSatoshis: 21,
            description: "sending to username"
        }
    }
    const sendBitcoinResponse = mocked.mockResponse()

    await sendBitcoinMiddleware.sendBitcoinToLightningAddress(
        request,
        sendBitcoinResponse,
        next
    )
    
    expect(sendBitcoinResponse.statusCode).toBe(201)
    expect(sendBitcoinResponse.jsonResponse.status).toBe("ok")
    expect(sendBitcoinResponse.jsonResponse.data.status).toBe("pending")
    expect(sendBitcoinResponse.jsonResponse.data.amountInSatoshis).toBe(21)
    expect(sendBitcoinResponse.jsonResponse.data.feeInSatoshis).toBe(0)
    expect(sendBitcoinResponse.jsonResponse.data.sourceId).toBe(content.alice.id)
    expect(sendBitcoinResponse.jsonResponse.data.destinationId).toBe(content.bob.id)
    expect(sendBitcoinResponse.jsonResponse.data.destinationParameter).toBe(`${bobPhoneNumber}@8333.mobi`)
});

test("Fail to send Bitcoin with insufficient balance", async () => {
    const content = await init();

    const request = {
        user: content.bob,
        body: {
            destinationParameter: carolLightningAddress,
            amountInSatoshis: 21,
            description: "sending to username"
        }
    }
    const sendBitcoinResponse = mocked.mockResponse()

    await sendBitcoinMiddleware.sendBitcoinToLightningAddress(
        request,
        sendBitcoinResponse,
        next
    )
    
    expect(sendBitcoinResponse.statusCode).toBe(413)
    expect(sendBitcoinResponse.jsonResponse.status).toBe("error")
    expect(sendBitcoinResponse.jsonResponse.message).toBe("Insufficient balance")
});

test("Send Bitcoin to machankura phoneNumber lightning address", async () => {
    const content = await init();

    const request = {
        user: content.alice,
        body: {
            destinationParameter: `satoshi@machankura.com`,
            amountInSatoshis: 21,
            description: "sending to lightning address"
        }
    }
    const sendBitcoinResponse = mocked.mockResponse()

    await sendBitcoinMiddleware.sendBitcoinToLightningAddress(
        request,
        sendBitcoinResponse,
        next
    )
    
    expect(sendBitcoinResponse.statusCode).toBe(201)
    expect(sendBitcoinResponse.jsonResponse.status).toBe("ok")
    expect(sendBitcoinResponse.jsonResponse.data.status).toBe("pending")
    expect(sendBitcoinResponse.jsonResponse.data.amountInSatoshis).toBe(21)
    expect(sendBitcoinResponse.jsonResponse.data.feeInSatoshis).toBe(1)
    expect(sendBitcoinResponse.jsonResponse.data.sourceId).toBe(content.alice.id)
    expect(sendBitcoinResponse.jsonResponse.data.destinationId).toBeUndefined()
    expect(sendBitcoinResponse.jsonResponse.data.destinationParameter).toBe(`satoshi@machankura.com`)
});
